# -*- coding: utf-8 -*-

"""Utilities for identification algorithms."""

from __future__ import annotations

from typing import Any, Iterable, Optional, Union

import networkx as nx

from y0.dsl import (
    CounterfactualVariable,
    Distribution,
    Expression,
    Intervention,
    P,
    Probability,
    Variable,
)
from y0.graph import NxMixedGraph
from y0.mutate.canonicalize_expr import canonical_expr_equal

__all__ = [
    "Query",
    "Identification",
    "Unidentifiable",
    "str_nodes_to_variable_nodes",
]


class Unidentifiable(Exception):  # noqa:N818
    """Raised on failure of the identification algorithm."""


class Query:
    """An identification query."""

    outcomes: set[Variable]
    treatments: set[Variable]
    conditions: set[Variable]

    def __init__(
        self,
        outcomes: set[Variable],
        treatments: set[Variable],
        conditions: Optional[set[Variable]] = None,
    ) -> None:
        """Instantiate an identification.

        :param outcomes: The outcomes in the query
        :param treatments: The treatments in the query (e.g., counterfactual variables)
        :param conditions: The conditions in the query (e.g., coming after the bar)
        :raises TypeError: If any of the outcomes, treatements, or conditions are not vanilla variables
        """
        if not all(isinstance(v, Variable) for v in outcomes):
            raise TypeError
        elif any(isinstance(v, CounterfactualVariable) for v in outcomes):
            raise TypeError
        else:
            self.outcomes = outcomes

        if not all(isinstance(v, Variable) for v in treatments):
            raise TypeError
        elif any(isinstance(v, CounterfactualVariable) for v in treatments):
            raise TypeError
        else:
            self.treatments = treatments

        if conditions is None:
            self.conditions = set()
        elif not all(isinstance(v, Variable) for v in conditions):
            raise TypeError
        elif any(isinstance(v, CounterfactualVariable) for v in conditions):
            raise TypeError
        else:
            self.conditions = conditions

    def __eq__(self, other: Any) -> bool:
        """Check if the outcomes, treatments, and conditions are equal."""
        return (
            isinstance(other, Query)
            and self.outcomes == other.outcomes
            and self.treatments == other.treatments
            and self.conditions == other.conditions
        )

    @classmethod
    def from_str(
        cls,
        outcomes: Union[str, Iterable[str]],
        treatments: Union[str, Iterable[str]],
        conditions: Optional[Iterable[str]] = None,
    ) -> Query:
        """Construct a query from text variable names."""
        return cls(
            outcomes={Variable(outcomes)}
            if isinstance(outcomes, str)
            else {Variable(n) for n in outcomes},
            treatments={Variable(treatments)}
            if isinstance(treatments, str)
            else {Variable(n) for n in treatments},
            conditions=None if conditions is None else {Variable(n) for n in conditions},
        )

    @classmethod
    def from_expression(
        cls,
        query: Union[Probability, Distribution],
    ) -> Query:
        """Instantiate an identification.

        :param query: The query probability expression
        :returns: An identification tuple
        :raises ValueError: If there are ragged counterfactual variables in the query
        """
        outcomes = {child.get_base() for child in query.children}  # clean counterfactuals
        conditions = {parent.get_base() for parent in query.parents}

        first_child = query.children[0]
        if not isinstance(first_child, CounterfactualVariable):
            if _unexp_interventions(query.children) or _unexp_interventions(query.parents):
                raise ValueError("Inconsistent usage of interventions")
            treatments = set()
        else:
            interventions = set(first_child.interventions)
            if _ragged_interventions(query.children, interventions) or _ragged_interventions(
                query.parents, interventions
            ):
                raise ValueError("Inconsistent usage of interventions")
            treatments = {intervention.get_base() for intervention in first_child.interventions}

        return Query(
            outcomes=outcomes,
            treatments=treatments,
            conditions=conditions,
        )

    def exchange_observation_with_action(
        self, variables: Union[Variable, Iterable[Variable]]
    ) -> Query:
        """Move the condition variable(s) to the treatments."""
        if isinstance(variables, Variable):
            variables = {variables}
        else:
            variables = set(variables)
        if any(v not in self.conditions for v in variables):
            raise ValueError
        return Query(
            outcomes=self.outcomes,
            treatments=self.treatments | variables,
            conditions=self.conditions - variables,
        )

    def exchange_action_with_observation(
        self, variables: Union[Variable, Iterable[Variable]]
    ) -> Query:
        """Move the treatment variable(s) to the conditions."""
        if isinstance(variables, Variable):
            variables = {variables}
        else:
            variables = set(variables)
        if any(v not in self.treatments for v in variables):
            raise ValueError
        return Query(
            outcomes=self.outcomes,
            treatments=self.treatments - variables,
            conditions=self.conditions | variables,
        )

    def with_treatments(self, extra_treatments: Iterable[Variable]) -> Query:
        """Create a new identification with additional treatments."""
        return Query(
            outcomes=self.outcomes,
            treatments=self.treatments.union(extra_treatments),
            conditions=self.conditions,
        )

    def uncondition(self) -> Query:
        """Move the conditions to outcomes."""
        return Query(
            outcomes=self.outcomes | self.conditions,
            treatments=self.treatments,
            conditions=None,
        )

    @property
    def expression(self) -> Expression:
        """Return the query as a Probabilistic expression."""
        if self.conditions and self.treatments:
            return P[self.treatments](self.outcomes | self.conditions)
        elif self.treatments:
            return P[self.treatments](self.outcomes)
        elif self.conditions:
            return P(self.outcomes | self.conditions)
        else:
            return P(self.outcomes)


def _unexp_interventions(variables: Iterable[Variable]) -> bool:
    return any(isinstance(c, CounterfactualVariable) for c in variables)


def _ragged_interventions(variables: Iterable[Variable], interventions: set[Intervention]) -> bool:
    return not all(
        isinstance(child, CounterfactualVariable) and set(child.interventions) == interventions
        for child in variables
    )


class Identification:
    """A package of a query and resulting estimand from identification on a graph."""

    query: Query
    graph: NxMixedGraph
    estimand: Expression

    def __init__(
        self,
        query: Query,
        graph: NxMixedGraph,
        estimand: Optional[Expression] = None,
    ) -> None:
        """Instantiate an identification.

        :param query: The generalizd identification query (outcomes/treatments/conditions)
        :param graph: The graph
        :param estimand: If none is given, will use the joint distribution over all variables in the graph.
        """
        self.query = query
        self.graph = str_nodes_to_variable_nodes(graph)
        self.estimand = P(self.graph.nodes()) if estimand is None else estimand

    @classmethod
    def from_parts(
        cls,
        outcomes: set[Variable],
        treatments: set[Variable],
        graph: NxMixedGraph,
        estimand: Optional[Expression] = None,
        conditions: Optional[set[Variable]] = None,
    ) -> Identification:
        """Instantiate an identification.

        :param outcomes: The outcomes in the query
        :param treatments: The treatments in the query (e.g., counterfactual variables)
        :param conditions: The conditions in the query (e.g., coming after the bar)
        :param graph: The graph
        :param estimand: If none is given, will use the joint distribution over all variables in the graph.
        :returns: An identification object
        """
        return cls(
            query=Query(outcomes=outcomes, treatments=treatments, conditions=conditions),
            graph=graph,
            estimand=estimand,
        )

    @classmethod
    def from_expression(
        cls,
        *,
        query: Union[Probability, Distribution],
        graph: NxMixedGraph,
        estimand: Optional[Expression] = None,
    ) -> Identification:
        """Instantiate an identification.

        :param query: The query probability expression
        :param graph: The graph
        :param estimand: If none is given, will use the joint distribution over all variables in the graph.
        :returns: An identification object
        """
        return cls(
            query=Query.from_expression(query),
            graph=graph,
            estimand=estimand,
        )

    @property
    def outcomes(self) -> set[Variable]:
        """Return this identification object's query's outcomes."""
        return self.query.outcomes

    @property
    def treatments(self) -> set[Variable]:
        """Return this identification object's query's treatments."""
        return self.query.treatments

    @property
    def conditions(self) -> set[Variable]:
        """Return this identification object's query's conditions."""
        return self.query.conditions

    def exchange_observation_with_action(
        self, variables: Union[Variable, Iterable[Variable]]
    ) -> Identification:
        """Move the condition variable(s) to the treatments."""
        return Identification(
            query=self.query.exchange_observation_with_action(variables),
            graph=self.graph,
            estimand=self.estimand,
        )

    def exchange_action_with_observation(
        self, variables: Union[Variable, Iterable[Variable]]
    ) -> Identification:
        """Move the treatment variable(s) to the conditions."""
        return Identification(
            query=self.query.exchange_action_with_observation(variables),
            graph=self.graph,
            estimand=self.estimand,
        )

    def with_treatments(self, extra_treatments: Iterable[Variable]) -> Identification:
        """Create a new identification with additional treatments."""
        return Identification(
            query=self.query.with_treatments(extra_treatments),
            estimand=self.estimand,
            graph=self.graph,
        )

    def uncondition(self) -> Identification:
        """Move the conditions to outcomes."""
        return Identification(
            query=self.query.uncondition(),
            estimand=self.estimand,
            graph=self.graph,
        )

    def __repr__(self) -> str:
        return (
            f'Identification(outcomes="{self.outcomes}, treatments="{self.treatments}",'
            f'conditions="{self.conditions}",  graph="{self.graph!r}", estimand="{self.estimand}")'
        )

    def __eq__(self, other: Any) -> bool:
        """Check if the query, estimand, and graph are equal."""
        return (
            isinstance(other, Identification)
            and self.query == other.query
            and canonical_expr_equal(self.estimand, other.estimand)
            and self.graph == other.graph
        )


def str_nodes_to_variable_nodes(graph: NxMixedGraph) -> NxMixedGraph:
    """Generate a variable graph from this graph of strings."""
    return NxMixedGraph.from_edges(
        nodes={Variable.norm(node) for node in graph.nodes()},
        directed=_convert(graph.directed),
        undirected=_convert(graph.undirected),
    )


def _convert(graph: nx.Graph) -> list[tuple[Variable, Variable]]:
    return [(Variable.norm(u), Variable.norm(v)) for u, v in graph.edges()]



"""Estimation of probabilities generated from identification."""

from __future__ import annotations

from contextlib import redirect_stdout
from typing import List, Optional, Union

import pandas as pd

from y0.algorithm.identify import identify_outcomes
from y0.dsl import CounterfactualVariable, Variable
from y0.graph import (
    NxMixedGraph,
    is_a_fixable,
    is_markov_blanket_shielded,
    is_p_fixable,
)

__all__ = [
    "estimate_ace",
]


def estimate_ace(
    graph: NxMixedGraph,
    treatments: Union[Variable, List[Variable]],
    outcomes: Union[Variable, List[Variable]],
    data: pd.DataFrame,
    *,
    conditions: Optional[List[Variable]] = None,
    bootstraps: int | None = None,
    alpha: float | None = None,
    estimator: Optional[str] = None,
) -> float:
    """Estimate the average treatment effect."""
    if conditions is not None:
        raise NotImplementedError("can not yet handle conditional queries")
    if isinstance(treatments, list) or isinstance(outcomes, list):
        raise NotImplementedError("can not yet handle multiple treatments nor outcomes")
    if isinstance(treatments, CounterfactualVariable) or isinstance(
        outcomes, CounterfactualVariable
    ):
        raise NotImplementedError("can not yet handle counterfactual treatments nor outcomes")
    if not df_covers_graph(graph=graph, data=data):
        raise ValueError

    # explicitly encode suggestions from Ananke
    if estimator is not None:
        pass
    elif is_a_fixable(graph, treatments):
        if is_markov_blanket_shielded(graph):
            estimator = "eff-aipw"
        else:
            estimator = "aipw"
    elif is_p_fixable(graph, treatments):
        if is_markov_blanket_shielded(graph):
            estimator = "eff-apipw"
        else:
            estimator = "apipw"
    elif identify_outcomes(graph=graph, treatments=treatments, outcomes=outcomes) is not None:
        estimator = "anipw"
    else:
        raise RuntimeError("Graph is not identifiable; effect can not be estimated")

    return ananke_average_causal_effect(
        graph=graph,
        data=data,
        treatment=treatments,
        outcome=outcomes,
        estimator=estimator,
        bootstraps=bootstraps,
        alpha=alpha,
    )


def df_covers_graph(graph: NxMixedGraph, data: pd.DataFrame) -> bool:
    """Check if all variables in the graph appear as columns in the dataframe."""
    if graph.is_counterfactual():
        raise NotImplementedError
    graph_names = {node.name for node in graph.nodes()}
    data_names = set(data.columns)
    return graph_names.issubset(data_names)


def ananke_average_causal_effect(
    graph: NxMixedGraph,
    data: pd.DataFrame,
    treatment: Variable,
    outcome: Variable,
    estimator: str,
    bootstraps: int | None = None,
    alpha: float | None = None,
) -> float:
    """Estimate the average causal effect using Ananke."""
    ananke_graph = graph.to_admg()
    from ananke.estimation import CausalEffect

    with redirect_stdout(None):
        # redirect stdout gets rid of the unnecessary printing from Ananke,
        # e.g., when CausalEffect says what estimators can be used. We take
        # care of that explicitly below
        causal_effect = CausalEffect(ananke_graph, treatment.name, outcome.name)

    return causal_effect.compute_effect(
        data, estimator=estimator, n_bootstraps=bootstraps or 0, alpha=alpha or 0.05
    )



# -*- coding: utf-8 -*-

"""Implementation of the ID* algorithm."""

import itertools as itt
import logging
from typing import Collection, FrozenSet, Iterable, Mapping, Set, Tuple, cast

from .cg import is_not_self_intervened, make_counterfactual_graph
from .utils import Unidentifiable
from ...dsl import (
    CounterfactualVariable,
    Event,
    Expression,
    Intervention,
    One,
    P,
    Probability,
    Product,
    Sum,
    Variable,
    Zero,
)
from ...graph import NxMixedGraph

__all__ = [
    "id_star",
]

District = FrozenSet[Variable]
DistrictInterventions = Mapping[District, Event]

logger = logging.getLogger(__name__)


def id_star(graph: NxMixedGraph, event: Event, *, _number_recursions: int = 0) -> Expression:
    """Apply the ``ID*`` algorithm to the graph."""
    logger.debug(
        "[%d]: Calling ID* algorithm with graph G with\n\t nodes: %s\n"
        "\t directed: %s\n\t undirected %s\n"
        "\t outcome event: %s",
        _number_recursions,
        graph.nodes(),
        graph.directed.edges(),
        graph.undirected.edges(),
        event,
    )
    # Line 1: There's nothing in the counterfactual event (i.e., an empty conjunction),
    # then its probability is 1, by convention.
    if not event:
        return One()
    # Line 2: This violates the Axiom of Effectiveness
    if violates_axiom_of_effectiveness(event):
        return Zero()
    # Line 3: This is a tautological event and can be removed without affecting the probability
    reduced_event = remove_event_tautologies(event)
    if reduced_event != event:
        logger.debug("[%d] recurring on reduced event %s", _number_recursions, reduced_event)
        return id_star(graph, reduced_event, _number_recursions=_number_recursions + 1)
    # Line 4: invokes make-cg to construct a counterfactual graph :math:`G'` , and the
    # corresponding relabeled counterfactual event.
    cf_graph, new_event = make_counterfactual_graph(graph, event)
    logger.debug(
        "[%d] ID* Returned from make_counterfactual_graph(). New event: %s\n"
        "\tcounterfactual graph:\n"
        "\t nodes: %s\n"
        "\t directed: %s\n"
        "\t undirected: %s",
        _number_recursions,
        new_event,
        cf_graph.nodes(),
        cf_graph.directed.edges(),
        cf_graph.undirected.edges(),
    )
    # Line 5:
    if new_event is None:
        return Zero()

    # Line 6:
    nodes = set(node for node in cf_graph.nodes() if is_not_self_intervened(node))
    cf_subgraph = cf_graph.subgraph(nodes)
    if not cf_subgraph.is_connected():
        summand, events_of_each_district = id_star_line_6(cf_graph, new_event)
        logger.debug("[%d] summand: %s", _number_recursions, summand)
        assert 1 < len(events_of_each_district)
        logger.debug(
            "[%d] recurring on each district: %s ", _number_recursions, events_of_each_district
        )
        return Sum.safe(
            Product.safe(
                id_star(graph, events_of_district, _number_recursions=_number_recursions + 1)
                for events_of_district in events_of_each_district.values()
            ),
            summand,
        )

    # Line 7:
    conflicts = get_conflicts(cf_subgraph, new_event)
    if conflicts:
        raise ConflictUnidentifiable(cf_subgraph, new_event, conflicts)

    # Line 9
    return id_star_line_9(cf_subgraph)


class ConflictUnidentifiable(Unidentifiable):  # noqa:N818
    """An exception raised when line 8 of the ID* algorithm determines that the query is not identifiable.

    This happens because the event contains a conflict, i.e., an inconsistent value
    assignment where at least one value is in the subscript.
    """

    def __init__(
        self,
        cf_graph: NxMixedGraph,
        event: Event,
        conflicts: list[tuple[Intervention, Intervention]],
    ):
        """Instantiate the exception."""
        self.cf_graph = cf_graph
        self.event = event
        self.conflicts = conflicts


def get_free_variables(cf_graph: NxMixedGraph, event: Event) -> Set[Variable]:
    """Get the possible values of the counterfactual variables in the graph that are "free".

    i.e. that don't have values fixed by the event or a self-intervention.

    :param cf_graph: A counterfactual graph
    :param event: a conjunction of counterfactual variables
    :returns: The set of free variables
    """
    free_variables = {variable for variable in cf_graph.nodes() if is_not_self_intervened(variable)}
    return {v.get_base() for v in free_variables} - {e.get_base() for e in event}


def violates_axiom_of_effectiveness(event: Event) -> bool:
    r"""Run line 2 of the ID* algorithm.

    The second line states that if :math:`\event` contains a counterfactual
    which violates the Axiom of Effectiveness (Pearl, 2000), then :math:`\event`
    is inconsistent, and we return probability 0.

    :param event: a conjunction of counterfactual variables
    :returns: True if violates axiom of effectiveness
    """
    return any(
        intervention.get_base() == value.get_base() and value.star != intervention.star
        for counterfactual, value in event.items()
        if isinstance(counterfactual, CounterfactualVariable)
        for intervention in counterfactual.interventions
    )


def remove_event_tautologies(event: Event) -> Event:
    r"""Run line 3 of the ID* algorithm.

    The third line states that if a counterfactual contains its own value in the subscript,
    then it is a tautological event, and it can be removed from :math:`\event` without
    affecting its probability.

    :param event: a conjunction of counterfactual variables
    :return: updated event or None
    """
    return {
        variable: value
        for variable, value in event.items()
        if not is_redundant_counterfactual(variable, value)
    }


def is_redundant_counterfactual(variable: Variable, value: Intervention) -> bool:
    """Check if a counterfactual variable is intervened on itself and has the same value as the intervention."""
    if not isinstance(variable, CounterfactualVariable):
        return False
    return any(
        intervention.get_base() == value.get_base() and value.star == intervention.star
        for intervention in variable.interventions
    )


def id_star_line_6(
    cf_graph: NxMixedGraph, event: Event
) -> Tuple[Collection[Variable], DistrictInterventions]:
    r"""Run line 6 of the ID* algorithm.

    Line 6 is analogous to Line 4 in the ID algorithm, it decomposes the problem into a
    set of subproblems, one for each C-component in the counterfactual graph. In the ID
    algorithm, the term corresponding to a given C-component :math:`S_i` of the causal
    diagram was the effect of all variables not in :math:`S_i` on variables in :math:`S_i` ,
    in other words :math:`P_{\mathbf{v}\backslash s_i} (s_i )`, and the outermost summation
    on line 4 was over values of variables not in :math:`\mathbf{Y},\mathbf{X}`. Here, the
    term corresponding to a given C-component :math:`S^i` of the counterfactual graph :math:`G'`
    is the conjunction of counterfactual variables where each variable contains in its
    subscript all variables not in the C-component :math:`S^i` , in other words
    :math:`\mathbf{v}(G' )\backslash s^i` , and the outermost summation is over observable
    variables not in :math:`\event'` , that is over :math:`\mathbf{v}(G' ) \backslash \event'` ,
    where we interpret :math:`\event'` as a set of counterfactuals, rather than a conjunction.

    :param cf_graph: an NxMixedGraph
    :param event: a conjunction of counterfactual variables
    :return: a set of Variables in summand, a dictionary of districts and events
    """
    # Then we intervene on the Markov pillow of each district
    summand = get_free_variables(cf_graph, event)
    events_of_each_district = get_events_of_each_district(cf_graph, event)
    return summand, events_of_each_district


def get_events_of_each_district(graph: NxMixedGraph, event: Event) -> DistrictInterventions:
    """For each district, intervene each node in the Markov pillow of the district.

    Self-interventions are not considered part of the district

    :param graph: an NxMixedGraph
    :param event: a conjunction of counterfactual variables
    :return: a dictionary of districts and interventions of districts
    """
    nodes = {node for node in graph.nodes() if is_not_self_intervened(node)}
    subgraph = graph.subgraph(nodes)
    return {
        district: get_events_of_district(graph, district, event)
        for district in subgraph.districts()
    }


def get_events_of_district(graph, district, event) -> Event:
    """Create new events by intervening each node on the Markov pillow of the district.

    If the node in in the original event, then the value of the new event is the same as the original event.

    :param graph: an NxMixedGraph
    :param district: a district of the graph
    :param event: a conjunction of counterfactual variables
    :return: the events of the district
    """
    markov_pillow = graph.get_markov_pillow(district)
    if not markov_pillow:
        return {node.get_base(): _get_node_event(node, event) for node in district}
    return {
        node.get_base().intervene(markov_pillow): _get_node_event(node, event) for node in district
    }


def _get_node_event(node: Variable, event: Event) -> Intervention:
    if node in event:
        return event[node]
    return cast(Intervention, -node.get_base())


def get_conflicts(cf_graph: NxMixedGraph, event: Event) -> list[tuple[Intervention, Intervention]]:
    r"""Identify conflicts between interventions int the graph and value assignments in the event.

    .. note:: This is part of line 8 of the ID* algorithm

    :param cf_graph: an NxMixedGraph
    :param event: a joint distribution over counterfactual variables
    :return: A list of pairs of conflicts. The same intervention or value assignment may appear in multiple conflicts
    """
    interventions = get_cf_interventions(cf_graph.nodes())
    evidence = get_evidence(event)
    return [
        (intervention, ev)
        for intervention, ev in itt.product(interventions, evidence)
        if intervention.name == ev.name and intervention.star != ev.star
    ]


def get_cf_interventions(nodes: Iterable[Variable]) -> Set[Intervention]:
    """For the graph, get the set of interventions in each counterfactual variable (all the subscripts).

    .. note:: This was called ``sub()`` in the paper

    :param nodes: An iterable of Variables, potentially containing CounterfactualVariables
    :returns: The set of interventions over all counterfactual variables
    """
    return {
        intervention
        for node in nodes
        if isinstance(node, CounterfactualVariable)
        for intervention in node.interventions
    }


def get_evidence(event: Event) -> Set[Intervention]:
    """Get the evidence (interventions and values) of the counterfactual conjunction.

    The evidence (either set or observed) appearing in a given counterfactual conjunction
    (or set of counterfactual events)

    .. note:: This was called ``ev()`` in the paper

    :param event: a conjunction of counterfactual variables
    :returns: The set of values and interventions in the given counterfactual conjuction
    """
    return set(event.values()) | get_cf_interventions(event)


def id_star_line_9(cf_graph: NxMixedGraph) -> Probability:
    r"""Run line 9 of the ID* algorithm.

    Line 9 says if there are no conflicts, then it's safe to take the union of all
    subscripts in :math:`\event'` , and return the effect of the subscripts in :math:`\event'`
    on the variables in :math:`\event'`.

    :param cf_graph: A counterfactual graph
    :return: An interventional distribution.
    """
    interventions = get_cf_interventions(cf_graph.nodes())
    bases = [node.get_base() for node in cf_graph.nodes()]
    if len(interventions) > 0:
        return P[interventions](bases)
    else:
        return P(bases)



# -*- coding: utf-8 -*-

"""Implementation of the IDC* algorithm."""

import logging
from typing import Iterable, Tuple

from .cg import is_not_self_intervened, make_counterfactual_graph
from .id_star import id_star
from .utils import Unidentifiable
from ..conditional_independencies import are_d_separated
from ...dsl import Event, Expression, Variable, Zero
from ...graph import NxMixedGraph

__all__ = [
    "idc_star",
]

logger = logging.getLogger(__name__)


def get_new_outcomes_and_conditions(
    new_event: Event, outcomes: Event, conditions: Event
) -> Tuple[Event, Event]:
    """Get the new outcomes and conditions."""
    remaining_outcomes, missing_outcomes = get_remaining_and_missing_events(new_event, outcomes)
    remaining_conditions, missing_conditions = get_remaining_and_missing_events(
        new_event, conditions
    )
    new_event_keys = set(new_event) - set(outcomes) - set(conditions)
    if len(missing_outcomes) > 0 and len(missing_conditions) > 0:
        for outcome in new_event_keys:
            if outcome.get_base() in {missing.get_base() for missing in missing_outcomes}:
                remaining_outcomes[outcome] = new_event[outcome]
        for condition in new_event_keys:
            if condition.get_base() in {missing.get_base() for missing in missing_conditions}:
                remaining_conditions[condition] = new_event[condition]
        return remaining_outcomes, remaining_conditions
    elif len(missing_outcomes) > 0:
        for outcome in new_event_keys:
            remaining_outcomes[outcome] = new_event[outcome]
        return remaining_outcomes, remaining_conditions
    elif len(missing_conditions) > 0:
        for condition in new_event_keys:
            remaining_conditions[condition] = new_event[condition]
        return remaining_outcomes, remaining_conditions
    else:
        return remaining_outcomes, remaining_conditions


def get_remaining_and_missing_events(new_event: Event, old_event: Event) -> Tuple[Event, Event]:
    """Get the outcome from the event."""
    remaining = {k: v for k, v in old_event.items() if k in new_event}
    missing = {k: v for k, v in old_event.items() if k not in new_event}
    return remaining, missing


def idc_star(
    graph: NxMixedGraph, outcomes: Event, conditions: Event, *, _number_recursions: int = 0
) -> Expression:
    r"""Run the IDC* algorithm.

    :param graph: The causal graph
    :param outcomes: The outcome events corresponds to :math:`\gamma`
    :param conditions: The condition events corresponds to :math:`\delta`
    :param _number_recursions: The number of times the algorithm has recurred
    :returns: An expression created by the :func:`idc_star` algorithm after simplifying the original query
    :raises ValueError: If ID* returns zero
    """
    logger.debug(
        f"[{_number_recursions}]: Calling IDC* algorithm with graph G with\n\t nodes: {graph.nodes()}\n\t directed: "
        f"{graph.directed.edges()}\n\t undirected {graph.undirected.edges()}\n\t outcomes: {outcomes}\n\t "
        f"and conditions: {conditions}"
    )
    try:
        logger.debug(
            "[%d]: line 1 IDC* algorithm: call ID* algorithm with events %s",
            _number_recursions,
            conditions,
        )
        if isinstance(id_star(graph, conditions), Zero):
            raise ValueError("The ID* algorithm returned 0, so IDC* cannot be applied.")
        else:
            logger.debug(
                "[%d]: line 1 IDC* algorithm: ID* algorithm is not inconsistent", _number_recursions
            )
    except Unidentifiable:
        logger.debug(
            "[%d]: line 1 IDC* algorithm: ID* algorithm is not identifiable, but is also not inconsistent.",
            _number_recursions,
        )

    _events = outcomes | conditions
    logger.debug(
        "[%d]: line 2 IDC* algorithm: make counterfactual graph. Events: %s",
        _number_recursions,
        _events,
    )
    cf_graph, new_events = make_counterfactual_graph(graph, _events)
    logger.debug(
        f"[{_number_recursions}]: IDC* returned from make_counterfactual_graph with New events: {new_events}\n"
        "\tcounterfactual graph:\n"
        f"\t nodes: {cf_graph.nodes()}\n"
        f"\t directed: {cf_graph.directed.edges()}\n"
        f"\t undirected: {cf_graph.undirected.edges()}"
    )
    if new_events is None:
        logger.debug(
            f"[{_number_recursions}]: line 3 IDC* algorithm: make_counterfactual_graph is inconsistent. Returning Zero."
        )
        return Zero()

    new_outcomes, new_conditions = get_new_outcomes_and_conditions(new_events, outcomes, conditions)
    logger.debug(
        f"[{_number_recursions}]: line 3 IDC* algorithm: make_counterfactual_graph is not inconsistent.\n"
        f"\tNew outcomes: {outcomes}\tNew conditions: {conditions}\n"
    )
    for condition in new_conditions:
        logger.debug(
            f"[{_number_recursions}]: line 4 IDC* algorithm: for each condition, check if rule 2 of do calculus applies"
        )
        if cf_rule_2_of_do_calculus_applies(cf_graph, new_outcomes, condition):
            logger.debug(
                f"\t[{_number_recursions}]: line 4 IDC* algorithm: rule 2 of do calculus applies:\n\t\t{outcomes} "
                f"""is D-separated from {condition} in G{"'"*(_number_recursions + 1)} ({condition}_bar)"""
            )
            new_outcomes = {
                outcome.intervene(condition)
                if condition in cf_graph.ancestors_inclusive(outcome)
                else outcome: value
                for outcome, value in new_outcomes.items()
            }
            new_conditions = {k: v for k, v in new_conditions.items() if k != condition}
            logger.debug(
                f"\t[{_number_recursions}]: line 4 IDC* algorithm: call IDC* algorithm on new outcomes {new_outcomes} "
                f"and new conditions {new_conditions}"
            )
            return idc_star(
                graph, new_outcomes, new_conditions, _number_recursions=_number_recursions + 1
            )
        else:
            logger.debug(
                f"\t[{_number_recursions}]: line 4 IDC* algorithm: rule 2 of do calculus "
                f"does not apply to condition {condition}"
            )
    logger.debug(
        f"[{_number_recursions}]: line 5 IDC* algorithm: Call ID* algorithm on new events {new_outcomes} "
        f"| {new_conditions} and old graph:\n"
        f"\t nodes: {graph.nodes()}\n"
        f"\t directed: {graph.directed.edges()}\n"
        f"\t undirected: {graph.undirected.edges()}"
    )
    id_star_estimand = id_star(
        graph, new_outcomes | new_conditions, _number_recursions=_number_recursions + 1
    )
    logger.debug(f"[{_number_recursions}]: Returned from ID* with estimand {id_star_estimand}")
    if len(conditions) == 0:
        return id_star_estimand

    idc_star_estimand = id_star_estimand.conditional([c.get_base() for c in conditions])
    logger.debug(f"[{_number_recursions}]: Returning from IDC* with estimand {idc_star_estimand}")
    return idc_star_estimand


def cf_rule_2_of_do_calculus_applies(
    cf_graph: NxMixedGraph, outcomes: Iterable[Variable], condition: Variable
) -> bool:
    r"""Check if Rule 2 of the Do-Calculus applies to the conditioned variable.

    Rule 2 is about Observations that are interchangeable with Interventions.
    This occurs when there is no backdoor path from the conditioned variable to the outcomes.
    Rule 2 operationalizes this criterion to say that if we remove the outgoing edge from
    the conditioned variable, then if the intervened node is D-separated from all outcomes,
    then the backdoor criterion has been satisfied.

    :param cf_graph: an NxMixedGraph
    :param outcomes: The outcomes to check
    :param condition: The condition to check
    :returns: If rule 2 applies, see below.

    If Rule 2 of the do calculus applies to the conditioned variable, then it can be converted to a do variable.

    .. math::

        \newcommand\ci{\perp\!\!\!\perp}
        \newcommand{\ubar}[1]{\underset{\bar{}}{#1}}
        \newcommand{\obar}[1]{\overset{\bar{}}{#1}}
        \text{if } (\exists Z \in \mathbf{Z})(\mathbf{Y} \ci Z | \mathbf{X}, \mathbf{Z}
        - \{Z\})_{G_{\bar{\mathbf{X}}\ubar{Z}}} \\
        \text{then } P(\mathbf{Y}|do(\mathbf{X}),\mathbf{Z}) = P(\mathbf Y|do(\mathbf X), do(Z), \mathbf{Z} - \{Z\})
    """
    #: also called "blocked nodes"
    conditions = {n for n in cf_graph.nodes() if not is_not_self_intervened(n)}
    graph_mod = cf_graph.remove_out_edges(condition)
    return all(
        are_d_separated(graph_mod, outcome, condition, conditions=conditions)
        for outcome in outcomes
    )
